# -*- coding: utf-8 -*-
"""

    Тесты шифрования интервалов времени

"""

from trivial_tools.datetime_tools.coding import (
    encode_digits, encode_letters, encode_period, decode_letters, decode_digits, decode_period
)


def test_encode_digits_unique():
    """
    Должен выдавать уникальный код времени для дня
    """
    codes = set()

    for sector in [1, 2, 3]:
        for day in range(1, 32):
            code = encode_digits(day, sector)
            codes.add(code)

    assert len(codes) == 31 * 3


def test_encode_letters_unique():
    """
    Должен выдавать три диапазона с общим числом уникальныз хначений не менее 24 * 24 = 576
    """
    sectors = set()
    codes = set()

    from collections import defaultdict
    di = defaultdict(int)
    for minute in range(1, 1441):
        sector, code = encode_letters(minute)
        sectors.add(sector)
        codes.add(code)
        di[code] += 1

    assert len(sectors) == 3
    assert {1, 2, 3} == sectors
    assert len(codes) == 1440 // 3


def test_encode_period_unique():
    """
    Должен выдавать уникальный код времени для месяца
    """
    codes = set()
    combinations = set()

    for day in range(1, 32):
        for minute in range(1, 1441):
            code = encode_period(day, minute)
            codes.add(code)
            combinations.add((day, minute))

    assert len(codes) == 31 * 1440
    assert len(combinations) == 31 * 1440


def test_decode_digits():
    """
    Должен получить обратно номер дня + сектор
    """
    assert decode_digits('01') == (1, 1)
    assert decode_digits('64') == (31, 2)
    assert decode_digits('87') == (21, 3)
    assert decode_digits('97') == (31, 3)
    assert decode_digits('40') == (7, 2)
    assert decode_digits('14') == (14, 1)


def test_decode_letters():
    """
    Должен получить обратно номер дня + сектор
    """
    assert decode_letters(1, 'AA') == 1
    assert decode_letters(3, 'ZZ') == 1440


def test_full():
    """
    Сквозная проверка работоспособности
    """
    for day in range(1, 32):
        for minute in range(1, 1441):
            code = encode_period(day, minute)
            out_day, out_minute = decode_period(code)
            assert day == out_day
            assert minute == out_minute



# -*- coding: utf-8 -*-
"""

    Набор инструментов для упрощённого кодирования времени

"""
# сторонние модули
from typing import Tuple

# модули проекта общие
from trivial_tools.special.special import fail


LETTERS = 'A B C D E F G H J K L M N P Q R S T U V W X Y Z'.split()


def encode_period(day_num: int, minute_num: int) -> str:
    """
    Зашифровать период

    Пример шифра: AA25

    Сначала идут две буквы, потом две цифры. Буква I не используется
    Нумерация дней и минут начинается с 1

    >>> encode_period(day_num=1, minute_num=1)
    'AA01'

    >>> encode_period(day_num=7, minute_num=800)
    'PH40'

    >>> encode_period(day_num=14, minute_num=37)
    'BN14'

    >>> encode_period(day_num=21, minute_num=1270)
    'NX87'

    >>> encode_period(day_num=31, minute_num=1440)
    'ZZ97'
    """
    if not (1 <= day_num <= 31) or not (1 <= minute_num <= 1440):
        fail(f'Неправильно указано время: день {day_num} [1, 31], минута {minute_num} [1, 1440]')

    sector, letters = encode_letters(minute_num)
    digits = encode_digits(day_num, sector)

    return letters + digits


def encode_letters(minute_num: int) -> Tuple[int, str]:
    """
    Получить буквенное начало

    >>> encode_letters(1)
    (1, 'AA')
    >>> encode_letters(1440)
    (3, 'ZZ')
    """
    base = minute_num - 1
    sector = base // 480 + 1

    value = minute_num % 480
    higher = (value - 1) // len(LETTERS)

    if higher > len(LETTERS) - 1:
        higher //= len(LETTERS)

    lower = base % len(LETTERS)
    code = LETTERS[higher] + LETTERS[lower]

    return sector, code


def encode_digits(day_num: int, sector: int) -> str:
    """
    Получить числовую концовку

    >>> encode_digits(1, 1)
    '01'

    >>> encode_digits(31, 2)
    '64'
    """
    return '{:02d}'.format(int(day_num + (sector - 1) * 33))


def decode_period(code: str) -> Tuple[int, int]:
    """
    Преобразовать период из кода в кортеж [день, минута]

    >>> decode_period('AA11')
    (11, 1)
    """
    if len(code) != 4:
        fail(f'Неправильно указан код периода: {code}')

    letters = code[0:2]
    digits = code[2:]
    day_num, sector = decode_digits(digits)
    minute = decode_letters(sector, letters)
    return day_num, minute


def decode_letters(sector: int, code: str) -> int:
    """
    Раскодировать буквенную часть кода

    >>> decode_letters(1, 'AA')
    1
    """
    min_value = (sector - 1) * 480 + 1
    max_value = sector * 480

    for i in range(min_value, max_value + 1):
        check_sector, check_code = encode_letters(i)
        if check_code == code:
            return i

    fail(f'Не удалось расшифровать код времени: {code}')


def decode_digits(code: str) -> Tuple[int, int]:
    """
    Раскодировать цифровую часть кода

    >>> decode_digits('64')
    (31, 2)
    """
    value = int(code)
    day_num = value % 33
    sector = value // 33 + 1
    return day_num, sector



# -*- coding: utf-8 -*-
"""

    Регистратор методов для JSON-RPC.
    Используется для централизованного управления всеми методами приложения.

"""
# встроенные модули
import json
from enum import Enum
from inspect import signature
from types import FunctionType
from typing import Optional, List, Callable, Dict, Sequence, Tuple, Any, Union

# сторонние модули
from trivial_tools.formatters.base import s_type

# словарь запроса
SingleRequest = Dict[str, Any]
GroupRequest = List[Dict[str, Any]]

# словарь ответа
SingleResponse = Tuple[SingleRequest, int]


class HTTP(Enum):
    """
    Коды ответов
    """
    OK: int = 200
    BAD_REQUEST: int = 400
    NOT_AUTHORIZED: int = 401
    METHOD_NOT_ALLOWED: int = 405
    NOT_ACCEPTABLE: int = 406
    INTERNAL_ERROR: int = 500


class JSONRPCMethodMaster:
    """
    Регистратор методов для JSON-RPC.
    Используется для централизованного управления всеми методами приложения.
    """
    def __init__(self, ignores: Sequence[str] = ('secret_key',)):
        """
        Инициация
        """
        self._ignores = ['return'] + list(ignores)
        self._methods: Dict[str, Callable] = {}

    def __contains__(self, item):
        """
        Проверить, имеется ли этот метод в записях.
        """
        return item in self._methods

    def register_method(self, func: FunctionType) -> FunctionType:
        """
        Зарегистрировать метод.
        """
        self._methods[func.__name__] = func
        return func

    def unknown_method(self, method_name: str) -> bool:
        """
        Этот метод нам неизвестен?
        """
        return method_name not in self._methods

    def check_signature(self, body: Union[SingleRequest, GroupRequest]) -> List[SingleResponse]:
        """
        Проверка полученных параметров на соответствие сигнатуре метода
        """
        if isinstance(body, dict):
            # одиночный запрос
            error_msg = self._check_signature(body)
            if error_msg:
                return [({'code': -32602, 'message': error_msg}, HTTP.NOT_ACCEPTABLE)]
            return [(body, HTTP.OK)]

        if isinstance(body, list):
            # групповой запрос
            response = []
            for sub_body in body:
                error_msg = self._check_signature(sub_body)
                if error_msg:
                    response.append(({'code': -32602, 'message': error_msg}, HTTP.BAD_REQUEST))
                else:
                    response.append((sub_body, HTTP.OK))

            return response

        # bad request
        bad_response = {'code': -32700, 'message': f'Неправильный формат запроса.'}
        return_code = HTTP.BAD_REQUEST
        return [(bad_response, return_code)]

    def _check_signature(self, body: SingleRequest) -> str:
        """
        Проверка полученных параметров на соответствие сигнатуре метода
        """
        method_name = body.get('method')
        parameters = body.get('params').keys()

        method: FunctionType = self._methods[method_name]

        sig = signature(method)

        valid_keys = {x for x in sig.parameters if x not in self._ignores}
        other_keys = {x for x in parameters if x not in self._ignores}

        response = 'Расхождение в аргументах метода: '

        if valid_keys != other_keys:
            insufficient = ', '.join(valid_keys - other_keys)
            excess = ', '.join(other_keys - valid_keys)

            if insufficient:
                response += 'не хватает аргументов ' + insufficient

                if excess:
                    response += ', '

            if excess:
                response += 'лишние аргументы ' + excess

            return response
        return ''

    def get_method(self, method_name: str) -> Optional[Callable]:
        """
        Получить метод
        """
        return self._methods.get(method_name)

    def get_method_names(self) -> List[str]:
        """
        Получить имена доступных методов
        """
        return sorted(self._methods.keys())

    @staticmethod
    def extract_json(request) -> Tuple[Union[SingleRequest, List[SingleRequest]], int]:
        """
        Безопасное извлечение JSON из запроса
        """
        try:
            request_body = request.json()
            return_code = HTTP.OK

        except json.JSONDecodeError as err:
            request_body = {'code': -32700, 'message': f'Неправильный формат запроса, {err}'}
            return_code = HTTP.BAD_REQUEST

        return request_body, return_code

    def check_request(self, body: Union[SingleRequest, GroupRequest]) -> List[SingleResponse]:
        """
        Проверить на предмет соответствия JSON RPC 2.0
        """
        result = []

        if isinstance(body, dict):
            response, status_code = self.check_dict(body)
            result.append((response, status_code))

        elif isinstance(body, list):
            for sub_body in body:
                response, status_code = self.check_dict(sub_body)
                result.append((response, status_code))
        else:
            response = {'code': -32700, 'message': f'Неправильный формат запроса, {s_type(body)}'}
            result.append((response, HTTP.BAD_REQUEST))

        return result

    def check_dict(self, body: SingleRequest) -> SingleResponse:
        """
        Проверить словарь запроса
        """
        version = body.get('jsonrpc')
        method = body.get('method')
        params = body.get('params')

        response = body
        status_code = HTTP.OK

        if version != '2.0':
            response = {'code': -32600, 'message': 'Поддерживается только JSON-RPC 2.0.'}
            status_code = HTTP.BAD_REQUEST

        elif method is None:
            response = {'code': -32601, 'message': 'Не указан метод запроса.'}
            status_code = HTTP.BAD_REQUEST

        elif method not in self._methods:
            response = {'code': -32601, 'message': f'Неизвестный метод: {method}.'}
            status_code = HTTP.METHOD_NOT_ALLOWED

        elif params is None:
            response = {'code': -32602, 'message': 'Не указаны аргументы вызова.'}
            status_code = HTTP.BAD_REQUEST

        elif params.get('secret_key') is None:
            response = {'code': -32602, 'message': 'Не предоставлен ключ авторизации.'}
            status_code = HTTP.NOT_AUTHORIZED

        elif isinstance(params, list):
            response = {'code': -32602, 'message': 'Позиционные аргументы не поддерживаеются.'}
            status_code = HTTP.NOT_ACCEPTABLE

        elif not isinstance(params, dict):
            response = {'code': -32602, 'message': f'Неправильно оформлены аргументы: {params!r}'}
            status_code = HTTP.NOT_ACCEPTABLE

        return response, status_code

    @staticmethod
    def authorize(body: SingleRequest, check_func: Callable) -> SingleResponse:
        """
        Проверить авторизацию внешней функцией
        """
        response, status_code = check_func(body)

        if status_code != HTTP.OK:
            response = {'code': -32602, 'message': f'Отказано в доступе.'}
            status_code = HTTP.NOT_AUTHORIZED

        return response, status_code

    def execute(self, body: SingleRequest) -> Any:
        """
        Исполнить запрос
        """
        method_name = body.get('method')
        params = body.get('params', {})
        method = self.get_method(method_name)

        # секретный ключ не нужен внутри методов
        params.pop('secret_key', None)

        # @@@@@@@@@@@@@@@@@@@@@@@@
        response = method(**params)
        # @@@@@@@@@@@@@@@@@@@@@@@@

        return response
